# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from os import environ


def pytest_addoption(parser):
    parser.addoption("--auth_connection_info", default=environ.get("AUTH_CONNECTION_INFO"))



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import jwt
import time

from unittest import TestCase, mock
from dicom_client.token import TokenCache


def make_get_token_func():
    count = 0

    def get_token_func():
        nonlocal count
        count += 1
        return jwt.encode({"exp": 1606360460.985007, "count": count}, key="secret")

    return get_token_func


class TokenCacheTest(TestCase):
    def setUp(self):
        self.token_cache = TokenCache(make_get_token_func())

    @mock.patch("time.time", mock.MagicMock(return_value=1606360480.985007))
    def test_expired_jwt(self):
        self.assertTrue(self.token_cache.should_refresh())
        token1 = self.token_cache.get_token()
        self.assertIsNotNone(token1)
        self.assertTrue(self.token_cache.should_refresh())
        token2 = self.token_cache.get_token()
        self.assertIsNotNone(token2)
        self.assertNotEqual(token1, token2)

    @mock.patch("time.time", mock.MagicMock(return_value=1606360400.985007))
    def test_expired_jwt_copy(self):
        token1 = self.token_cache.get_token()
        self.assertIsNotNone(token1)
        self.assertFalse(self.token_cache.should_refresh())
        token2 = self.token_cache.get_token()
        self.assertIsNotNone(token2)
        self.assertEqual(token1, token2)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

""" A series of integration tests """

from pathlib import Path
import json
import logging
import os
import ast
import pydicom
from unittest import TestCase
import requests

from dicom_client import (
    DicomClient,
    TokenCache,
    make_get_token_func,
)

base_url = "https://tonydicom.azurewebsites.net"


class TestsIntegration(TestCase):
    """Testing some basic methods"""

    # -----------------------------------------
    # Set up some defaults for the tests
    # Specify input and output folders for DCM files
    # Specify the URL for the DICOM server
    def setUp(self):
        """ Set up the folder with DCM files """

        self._auth = ast.literal_eval(os.environ["AUTH_CONNECTION_INFO"])
        self._base_url = base_url
        self._output_folder = "./tests/sample_media/output"
        self._input_folder = "./tests/sample_media/input"
        self._input_filename = self._input_folder + "/" + "361_143_100_652.dcm"
        file_metadata = pydicom.filereader.dcmread(self._input_filename)
        study_id = file_metadata.StudyInstanceUID
        self._study_id_for_input_filename = study_id

    # -----------------------------------------
    # Test the ability to delete, upload, and download dcm files
    # Validates by checking a successfully downloaded dcm file
    def test_delete_upload_download_file(self):

        # Testing the upload capability means verfying that files are there and
        # have been uploaded correctly.

        # You can only do that be deleting the test data and then re-inserting it.
        # After re-insert, then download to make sure they are REALLY there.

        # After upload, verify you can download.

        # Get credentials
        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)

        # Start with an emtpy output folder
        self.clean_output_folder()

        # Do an upload
        try:
            self.test_delete_study_ids()
            dicom_client.upload_dicom_file(self._input_filename)
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None
        except Exception as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        # Verify a download is there and available.
        # This code should succeed because we just uploaded a file
        try:
            dicom_client.download_dicom(self._output_folder, self._study_id_for_input_filename)
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        # Verify we have only one file
        files = list(Path(self._output_folder).rglob("*.dcm"))
        number_files = len(files)
        # It really is the final number that matters for everything to work!
        assert number_files == 1

    # -----------------------------------------
    # Validate that we can get an oauth_token
    # Assert token is not empty
    def test_oauth_token(self):
        token_cache = TokenCache(make_get_token_func(self._auth))
        oauth_token = token_cache.get_token()
        assert not oauth_token is None

    # -----------------------------------------
    # Validate that we can instantiate DicomClient
    def test_dicom_client(self):
        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)
        assert not dicom_client is None

    # -----------------------------------------
    #  Validates that we can delete, upload, and retrieve
    #  DCM files using a study ID
    def test_query_by_patient_id(self):

        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)

        self.clean_output_folder()

        self.test_delete_study_ids()

        # Do an upload now
        dicom_client.upload_dicom_folder(self._input_folder)

        # /studies?PatientID=11235813
        study_ids = dicom_client.get_patient_study_ids("ID00042637202184406822975")
        assert "2.25.263916267626722293569390930007441210" == study_ids[0]

    # -----------------------------------------
    # Validates correct response to incorrectly formatted DCM file
    # Demonstrates the use of finding failure "reasons"
    def test_upload_bad_file(self):

        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)

        self.clean_output_folder()

        self.test_delete_study_ids()

        # Intentionally try to add a improperly formatted dcm file
        try:
            dicom_client.upload_dicom_file("./tests/sample_media/input/bad_dcm_file/temp.dcm")
        except requests.exceptions.HTTPError as errh:
            # There are more detailed errror codes embedded as JSON in response.text
            # See http://dicom.nema.org/medical/dicom/current/output/chtml/part18/sect_I.2.2.html
            logging.error("Http Error: %s", errh)
            assert errh.response.status_code == 409
            assert (
                json.loads(errh.response.text)["00081198"]["Value"][0]["00081197"]["Value"][0]
                == 43264
            )
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None
        except Exception as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

    # -----------------------------------------
    # A test designed to illustrate how to delete cdm files from Dicom Server
    # Intentionally try to download deleted file to demonstrate HTTP response 404
    def test_delete_file(self):

        # Setup so we have one file uploaded. Let's make sure we can delete it using study_id
        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)

        try:
            self.test_delete_study_ids()
            dicom_client.upload_dicom_file(self._input_filename)
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None
        except Exception as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        self.clean_output_folder()

        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)
        try:
            dicom_client.delete_dicom(self._study_id_for_input_filename)
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        # This attempted download should fail, resulting in a 404
        try:
            dicom_client.download_dicom(self._output_folder, self._study_id_for_input_filename)
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            # We better fail at the download because we deleted the dcm file
            assert errh.response.status_code == 404
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        # We have deleted the output folder. We should still have 0
        # as the download should fail.
        files = list(Path(self._output_folder).rglob("*.dcm"))
        number_files = len(files)
        # It really is the final number that matters for everything to work!
        assert number_files == 0

    # -----------------------------------------
    # Demonstrates how you can download multiple dcm files
    # bypassing in a single study ID
    def test_delete_upload_download(self):

        # Get some credentials
        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)

        self.clean_output_folder()

        self.test_delete_study_ids()

        # Do an upload now
        dicom_client.upload_dicom_folder(self._input_folder)

        # Verify a download is there.
        # No try/except needed because we know these study ids are in the DICOM server
        dicom_client.download_dicom(
            self._output_folder, "2.25.263916267626722293569390930007441210"
        )
        dicom_client.download_dicom(
            self._output_folder, "2.25.185320577304339382525212889304442143"
        )
        files = list(Path(self._output_folder).rglob("*.dcm"))
        number_files = len(files)
        # It really is the final number that matters for everything to work!
        assert number_files == 9

    # =============================================
    # Helper Methods
    # =============================================

    # -----------------------------------------
    # A test to demonstrate unexpected delete failures
    # This code demonstrates how to implement a delete functionality and deal with http 404s
    def test_delete_study_ids(self):
        """ Hard Coded Test that we can delete matching DCM files from DICOM server """
        token_cache = TokenCache(make_get_token_func(self._auth))
        dicom_client = DicomClient(self._base_url, token_cache)

        try:
            dicom_client.delete_dicom(
                "1.3.6.1.4.1.14519.5.2.1.6834.5010.465205689126985052184293614571"
            )
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            # For error 404, don't re-raise.
            # No expectation that dcm being deleted is there already.
            if errh.response.status_code != 404:
                raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        try:
            dicom_client.delete_dicom("2.25.263916267626722293569390930007441210")
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            # For error 404, don't re-raise.
            # No expectation that dcm being deleted is there already.
            if errh.response.status_code != 404:
                raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        try:
            dicom_client.delete_dicom("2.25.185320577304339382525212889304442143")
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            # For error 404, don't re-raise.
            # No expectation that dcm being deleted is there already.
            if errh.response.status_code != 404:
                raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        try:
            dicom_client.delete_dicom("1.2.276.0.50.192168001092.11517584.14547392.4")
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            # For error 404, don't re-raise.
            # No expectation that dcm being deleted is there already.
            if errh.response.status_code != 404:
                raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None
        return True

    # -----------------------------------------
    # A helper method to simply clear out the output folder
    def clean_output_folder(self):
        """ Remove DCM files on disk from a folder """
        files = list(Path(self._output_folder).rglob("*.dcm"))
        for file in files:
            os.remove(file)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""The main DicomClient object"""
from io import BytesIO
from pathlib import Path
import errno as _errno
import hashlib
import logging
import os
import pathlib
import pydicom
from urllib3.filepost import encode_multipart_formdata, choose_boundary
from abc import ABC, abstractmethod
import requests
import requests_toolbelt as tb


class DicomClientInterface(ABC):
    """Abstract base class that represents the interface to DicomClient()"""

    @abstractmethod
    def __init__(self, base_url, token_cache):
        """Constructor providing DICOM server URL and token cache object"""

    @abstractmethod
    def upload_dicom_folder(self, folder_name):
        """Upload DCM files from folder"""

    @abstractmethod
    def upload_dicom_file(self, file_name):
        """Upload a single DCM file"""

    @abstractmethod
    def delete_dicom(self, study_id, series_id=None, instance_id=None):
        """Delete DCM files and DICOM server"""

    @abstractmethod
    def download_dicom(self, output_folder, study_id, series_id=None, instance_id=None):
        """Download DCM files from DICOM server"""

    @abstractmethod
    def get_patient_study_ids(self, patient_id):
        """Get a collection of study IDs based on patient ID"""


class DicomClient(DicomClientInterface):

    """Supports upload, download, delete of DCM Files"""

    # ===========================================================
    def __init__(self, base_url, token_cache):
        """ Initialize """
        super().__init__(base_url, token_cache)
        self._base_url = base_url
        self._token_cache = token_cache

    # -------------------------------------------------
    # Core methods
    # -------------------------------------------------
    def upload_dicom_folder(self, folder_name):
        """Upload DCM files from a folder"""
        # Validate folder exists and has files
        self.validate_folder_name(folder_name)
        # Upload folder
        response = self.do_upload_dicom_folder(folder_name)
        return response

    def upload_dicom_file(self, file_name):
        """ Upload a single DCM file """
        self.validate_file_name(file_name)
        response = self.do_upload_dicom_file(file_name)
        return response

    def delete_dicom(self, study_id, series_id=None, instance_id=None):
        """ Delete Dicom Entries """
        response = self.do_delete_dicom(study_id, series_id, instance_id)
        return response

    def download_dicom(self, output_folder, study_id, series_id=None, instance_id=None):
        """ Download DCM Files """
        self.validate_output_folder(output_folder)
        response, _ = self.do_download_dicom(output_folder, study_id, series_id, instance_id)
        return response

    def get_patient_study_ids(self, patient_id):
        """ Download DCM Files """
        patient_id = "ID00042637202184406822975"
        response = self.do_get_patient_study_ids(patient_id)
        # 0020000D is the reference to the array of study ids for a patient id
        return response.json()[0]["0020000D"]["Value"]

    # -------------------------------------------------
    # Validations
    # -------------------------------------------------
    def validate_output_folder(self, output_folder):
        # Make sure folder exists
        if not os.path.isdir(output_folder):
            raise FileNotFoundError(
                _errno.ENOENT, "Upload folder not found = {}".format(output_folder)
            ) from None

    # -------------------------------------------------
    def validate_folder_name(self, folder_name):
        # Make sure folder exists
        if not os.path.isdir(folder_name):
            raise FileNotFoundError(
                _errno.ENOENT, "Upload folder not found = {}".format(folder_name)
            ) from None
        # Make sure files in folder exists
        files = list(Path(folder_name).rglob("*.dcm"))
        if len(files) == 0:
            raise FileNotFoundError(
                _errno.ENOENT, "No files in folder = {}".format(folder_name)
            ) from None

    # -------------------------------------------------
    def validate_file_name(self, file_name):
        """Validate that local file name exists"""
        if not os.path.isfile(file_name):
            raise FileNotFoundError(
                _errno.ENOENT, "Upload file not found = {}".format(file_name)
            ) from None

    # -------------------------------------------------
    # Supporting methods: OAUTH and Tokens
    # -------------------------------------------------
    def get_token(self):
        """Retrieve a wealth token from ADAL"""
        return self._token_cache.get_token()

    # -------------------------------------------------
    # Supporting methods: HTTP request system
    # -------------------------------------------------
    def do_download_dicom(self, output_folder, study_id, series_id=None, instance_id=None):
        """Download DCM files from DICOM server to output folder"""
        token = self.get_token()
        url = None
        if series_id is None and instance_id is None:
            url = f"/studies/{study_id}"
        if not series_id is None and instance_id is None:
            url = f"/studies/{study_id}/series/{series_id}"

        if not series_id is None and not instance_id is None:
            url = f"/studies/{study_id}/series/{series_id}/instances/{instance_id}"

        requests_session = requests.session()

        url = f"{self._base_url}{url}"

        headers = {
            "Accept": 'multipart/related; type="application/dicom"; transfer-syntax=*',
            "Authorization": "Bearer " + token,
        }

        response = None
        try:
            response = requests_session.get(url, headers=headers)
            response.raise_for_status()
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        mpd = tb.MultipartDecoder.from_response(response)
        self.write_files(output_folder, mpd)
        return response, len(mpd.parts)

    def do_upload_dicom_file(self, file_name):
        """Upload DCM file to DICOM server"""
        # Upload files and folders
        # Get filenames
        filenames = [Path(file_name)]
        response = self.upload_files(filenames)
        return response

    # -------------------------------------------------
    def do_upload_dicom_folder(self, folder_name):
        """Upload all DCM files and folder to DICOM server"""
        # Upload files and folders
        # Get filenames
        filenames = list(Path(folder_name).rglob("*.dcm"))
        response = self.upload_files(filenames)
        return response

    # -------------------------------------------------
    def do_get_patient_study_ids(self, patient_id):
        """Get collection of study IDs based on patient ID"""
        token = self.get_token()

        # The url that looks up by PatientID
        url = f"{self._base_url}/studies?PatientID={patient_id}"

        headers = {"Accept": "application/dicom+json", "Authorization": "Bearer " + token}

        requests_session = requests.session()

        response = None
        try:
            response = requests_session.get(url, headers=headers)
            response.raise_for_status()
            return response
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None
        return response

    # -------------------------------------------------
    def upload_files(self, filenames):
        """Upload DCM files using array of filenames"""
        token = self.get_token()
        # Get file content
        array_of_filecontent = []
        for file in filenames:
            with open(file, mode="rb") as file:  # b is important -> binary
                file_content = file.read()
            array_of_filecontent.append(file_content)
        # Build dictionary of file content to upload later
        dicom_files = {}
        for i in range(0, len(filenames)):
            dicom_files.update(
                {
                    filenames[i].stem: (
                        "dicomfile",
                        array_of_filecontent[i],
                        "application/dicom",
                    )
                }
            )
        # Encode into body
        boundary = choose_boundary()
        # Encode into body
        body, _ = encode_multipart_formdata(dicom_files, boundary)
        content_type = str("multipart/related; boundary=%s" % boundary)
        headers = {
            "Accept": "application/dicom+json",
            "Authorization": "Bearer " + token,
            "Content-Type": content_type,
        }
        url = f"{self._base_url}/studies"
        requests_session = requests.session()
        response = None
        try:
            response = requests_session.post(url, body, headers=headers, verify=False)
            response.raise_for_status()
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None
        return response

    def delete_dicom_file(self, study_id) -> bool:
        """Delete DCM files based on study ID"""
        oauth_token = self._token_cache.get_token()
        session = requests.session()

        url = f"{self._base_url}/studies/{study_id}"
        response = None
        try:
            headers = {
                "Accept": "application/dicom+json",
                "Authorization": "Bearer " + oauth_token,
            }
            response = session.delete(url, headers=headers, timeout=3)
            assert response != None
            return True
        except Exception:
            return False

    def do_delete_dicom(self, study_id, series_id=None, instance_id=None):
        """Delete DCM files based on combination of study ID, series ID, and instance ID"""
        token = self.get_token()
        url = None
        if series_id is None and instance_id is None:
            url = f"/studies/{study_id}"
        if not series_id is None and instance_id is None:
            url = f"/studies/{study_id}/series/{series_id}"

        if not series_id is None and not instance_id is None:
            url = f"/studies/{study_id}/series/{series_id}/instances/{instance_id}"

        requests_session = requests.session()

        url = f"{self._base_url}{url}"

        headers = {"Accept": "application/dicom+json", "Authorization": "Bearer " + token}

        response = None
        try:
            response = requests_session.delete(url, headers=headers, timeout=3)
            response.raise_for_status()
        except requests.exceptions.HTTPError as errh:
            logging.error("Http Error: %s", errh)
            raise errh from None
        except requests.exceptions.ConnectionError as errc:
            logging.error("Error Connecting: %s", errc)
            raise errc from None
        except requests.exceptions.Timeout as errt:
            logging.error("Timeout Error: %s", errt)
            raise errt from None
        except requests.exceptions.RequestException as err:
            logging.error("Unknown error type: %s", err)
            raise err from None

        return response

    # -------------------------------------------------
    # File System
    # -------------------------------------------------
    def write_files(self, folder, mpd):
        """ Write files from multi-part form """
        filenames = []
        for part in mpd.parts:
            filenames.append(self.write_content(folder, part.content))
        return filenames

    def write_single_file(self, folder, content):
        """ Write single file to disk """
        filenames = []
        filenames.append(self.write_content(folder, content))
        return filenames

    def write_content(self, folder, content):
        """ Write file to disk """
        dcm = pydicom.dcmread(BytesIO(content))
        patient_id = self.right(dcm.PatientID, 3)
        study_id = self.right(dcm.StudyInstanceUID, 3)
        series_id = self.right(dcm.SeriesInstanceUID, 3)
        instance_id = self.right(dcm.SOPInstanceUID, 3)
        filename = (
            folder
            + "/"
            + patient_id
            + "_"
            + study_id
            + "_"
            + series_id
            + "_"
            + instance_id
            + ".dcm"
        )
        # if I want the dcm in bytes
        dcm = pydicom.dcmread(BytesIO(content))
        # if file exists, and different contents, new filename and write
        # if file exists, and same contents, ignore
        if self.file_exists(filename):
            newfn = self.nextnonexistent(filename)
            pydicom.dcmwrite(newfn, dcm)
            if not self.different_contents(newfn, filename):
                os.remove(newfn)
        else:
            pydicom.dcmwrite(filename, dcm)
        return filename

    # ===========================================================
    @staticmethod
    def file_exists(file):
        """ Check if file exists """
        path = pathlib.Path(file)
        if path.is_file():
            return True
        return False

    @staticmethod
    def get_file_digest(filename):
        """ Get hash for file for duplicate check """
        with open(filename, "rb") as file:
            file_hash = hashlib.md5()
            chunk = file.read()
            file_hash.update(chunk)
        return file_hash.digest()

    def different_contents(self, newfn, filename):
        """ Check if file is different """
        return self.get_file_digest(newfn) != self.get_file_digest(filename)

    @staticmethod
    def nextnonexistent(filename):
        """Finds a unique filename in case of conflict with existing file"""
        fnew = filename
        root, ext = os.path.splitext(filename)
        i = 0
        while os.path.exists(fnew):
            i += 1
            fnew = "%s_%i%s" % (root, i, ext)
        return fnew

    # -------------------------------------------------
    @staticmethod
    def left(sourcestring, amount):
        """ Extract left most characters """
        return sourcestring[:amount]

    # -------------------------------------------------
    @staticmethod
    def right(sourcestring, amount):
        """ Extract right most characters """
        return sourcestring[-amount:]

    # -------------------------------------------------
    @staticmethod
    def mid(sourcestring, offset, amount):
        """ Extract middle chars """
        return sourcestring[offset : offset + amount]



from threading import Lock
import jwt
import adal
import time


class TokenRefreshError(Exception):
    pass


class TokenCache(object):
    """Manages adding tokens to the request header, as well as refreshing expired tokens"""

    DELTA = 10  # amount of seconds prior to expiration to refresh the token

    # get_token is simply a function that returns an access token.
    def __init__(self, get_token_func):
        """Constructor for TokenCache"""
        self.get_token_func = get_token_func
        self.token = None
        self.exp = None
        self._lock = Lock()

    def get_token(self):
        """Retrieve an oauth token"""
        if self.should_refresh():
            self.refresh()
        return self.token

    def should_refresh(self):
        """Check to see if oauth token needs refreshing"""
        return self.token is None or self.exp is None or self.exp <= time.time() + self.DELTA

    def refresh(self):
        """Retrieve a new oauth token"""
        with self._lock:
            self.token = self.get_token_func()
            self.exp = jwt.decode(self.token, verify=False)["exp"]
            if self.token is None:
                raise TokenRefreshError


def make_get_token_func(args):
    """Setup authentication context to enable token creation"""
    context = adal.AuthenticationContext(args["authority"])
    resource = args["oauth_resource"]
    client_secret = args["client_secret"]
    client_id = args["client_id"]

    def get_token_func():
        """Acquire token using authentication context"""
        resp = context.acquire_token_with_client_credentials(resource, client_id, client_secret)
        # print(resp)
        return resp["accessToken"]

    return get_token_func



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import setuptools
import os

with open("README_PYPI.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


tag_suffix = ""
if os.environ.get("DEV_VERSION_TAG") != None:
    tag_suffix = ".dev" + os.environ.get("DEV_VERSION_TAG")

setuptools.setup(
    name="dicom-client-python",  # Replace with your own username
    version="0.3.0" + tag_suffix,
    author="Braveheart",
    author_email="bterkaly@microsoft.com",
    description="A small dicom client in python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/microsoft/dicom-client-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "pydicom==2.1.1",
        "requests_toolbelt==0.9.1",
        "pyodbc==4.0.30",
        "urllib3==1.26.2",
        "requests==2.25.0",
        "adal==1.2.5",
        "PyJWT==1.7.1",
    ],
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

""" Utility functions """
import os
from pathlib import Path
import pydicom

# -------------------------------------------------
def left(sourcestring, amount):
    """ Extract the leftmost characters """
    return sourcestring[:amount]


# -------------------------------------------------
def right(sourcestring, amount):
    """ Extract the rightmost characters """
    return sourcestring[-amount:]


# -------------------------------------------------
def mid(sourcestring, offset, amount):
    """ Extracts some middle characters """
    return sourcestring[offset : offset + amount]


# -------------------------------------------------
def rename_files():
    """ Rename files based on metadata """
    files = list(Path(".").rglob("./dcmfiles3/*.dcm"))
    for file in files:
        metadata = pydicom.filereader.dcmread(file)
        patient_id = right(metadata.PatientID, 4)
        study_id = right(metadata.StudyInstanceUID, 4)
        series_id = right(metadata.SeriesInstanceUID, 4)
        instance_id = right(metadata.SOPInstanceUID, 4)
        newfn = (
            os.path.dirname(file)
            + "/"
            + patient_id
            + "_"
            + study_id
            + "_"
            + series_id
            + "_"
            + instance_id
            + ".dcm"
        )
        os.rename(file, newfn)


# -------------------------------------------------
def showdata():
    """ Display DCM metadata"""
    files = list(Path(".").rglob("./dcmfiles3/*.dcm"))
    for file in files:
        print(file)
        metadata = pydicom.filereader.dcmread(file)
        print("Patient ID = " + metadata.PatientID)
        print("Study ID = " + metadata.StudyInstanceUID)
        print("Series ID = " + metadata.SeriesInstanceUID)
        print("SOP Instance  ID = " + metadata.SOPInstanceUID)
        print("=========================================")


# -------------------------------------------------
def writecsv():
    """ Right out metadata to separated list """
    files = list(Path(".").rglob("./dcmfiles3/*.dcm"))
    with open("/mnt/c/temp/csvfile.csv", "w") as file:
        file.write("Patient, Study, Series, Instance\n")
        for file in files:
            metadata = pydicom.filereader.dcmread(file)
            result = (
                metadata.PatientID
                + ","
                + metadata.StudyInstanceUID
                + ","
                + metadata.SeriesInstanceUID
                + ","
                + metadata.SOPInstanceUID
                + "\n"
            )
            file.write(result)
